package table

import function.LispFunction
import sexpression.Cons
import sexpression.SExpression

import java.util.Stack

import sexpression.Nil.NIL

object ExecutionContext {

    private val functionCalls: Stack<LispFunctionRecurInfo> = Stack()

    var scope: SymbolTable = SymbolTable(NullSymbolTable)
    var isRecur: Boolean = false
        private set

    val isInFunctionCall: Boolean
        get() = !functionCalls.empty()

    val currentFunction: LispFunction
        get() = functionCalls.peek().lispFunction

    val isRecurInitializing: Boolean
        get() = functionCalls.peek().isRecurInitializing

    fun clearContext() {
        scope = SymbolTable(NullSymbolTable)
        functionCalls.clear()
        clearRecur()
    }

    fun restoreGlobalScope() {
        while (!scope.isGlobal)
            scope = scope.parent
    }

    fun lookupSymbolValue(symbolName: String): SExpression? {
        var t = scope
        while (t !== NullSymbolTable) {
            if (t.contains(symbolName))
                return t.get(symbolName)

            t = t.parent
        }

        return null
    }

    fun toList(): Cons {
        var symbols: Cons = NIL

        var t = scope
        while (t !== NullSymbolTable) {
            symbols = Cons(t.toList(), symbols)
            t = t.parent
        }

        return symbols
    }

    fun pushFunctionCall(function: LispFunction) {
        functionCalls.push(LispFunctionRecurInfo(function))
    }

    fun popFunctionCall() {
        functionCalls.pop()
    }

    fun setRecur() {
        isRecur = true
    }

    fun clearRecur() {
        isRecur = false
    }

    fun setRecurInitializing() {
        functionCalls.peek().setRecurInitializing()
    }

    fun clearRecurInitializing() {
        functionCalls.peek().clearRecurInitializing()
    }

    class LispFunctionRecurInfo(val lispFunction: LispFunction) {
        var isRecurInitializing: Boolean = false
            private set

        fun setRecurInitializing() {
            isRecurInitializing = true
        }

        fun clearRecurInitializing() {
            isRecurInitializing = false
        }
    }

    object NullSymbolTable : SymbolTable()
}
